// (c) Robert Kloefkorn 2010 
#include "gitter_hexa_top_pll.h"

// avoid linker errors by quadratures 
//#define MAPP_CUBE_3D_H_INCLUDED
#include "../serial/gitter_hexa_top.cc"

template < class A, class X, class MX >
void Hbnd4PllInternal < A, X, MX > :: HbndPll ::
setGhost ( const ghostpair_STI & gpair )
{
  if(gpair.first)
  {
    _ghostPair = gpair; 
    assert( _ghostPair.first );
    // copy indices from internal boundry to myhface3(.) of ghost
    _ghostPair.first->setIndicesAndBndId( *this->myhface4(0), _ghostPair.second );
  }
  else 
  {
    _ghostPair.first  =  0 ;
    _ghostPair.second = -1 ;
  }
}

template < class A, class X, class MX >
void Hbnd4PllInternal < A, X, MX > :: HbndPll ::  
splitGhost ( GhostChildrenInfo_t & info )
{
  if(_ghostPair.first)
  {
    typedef typename Gitter :: Geometric :: hexa_GEO  hexa_GEO;
    typedef typename Gitter :: Geometric :: hface4_GEO hface4_GEO;

    hexa_GEO & ghost = static_cast<hexa_GEO &> (*(_ghostPair.first)); 
    if(!ghost.down())
    {
      ghost.tagForGlobalRefinement();
      ghost.refine();
    } 

    int gFaceNum = _ghostPair.second;
    assert( gFaceNum >= 0 );
    assert( gFaceNum < 6 );

    hface4_GEO * face = ghost.myhface4( gFaceNum );
    assert( face );

    int count = 0; 
    for(face = face->down(); face; face = face->next() )
    {
      assert(face);
      hexa_GEO * ghch = 0; 

      typedef pair < Gitter :: Geometric :: hasFace4 *, int > neigh_t;
      neigh_t neighbour = face->nb.front();

      if( ! neighbour.first->isboundary ())
      {
        ghch = dynamic_cast<hexa_GEO *> (neighbour.first);
        assert(ghch);
        assert( ghch->up() == &ghost );
      }
      else
      {
        neighbour = face->nb.rear();
        assert( ! neighbour.first->isboundary () );
        ghch = dynamic_cast<hexa_GEO *> (neighbour.first);
      }

      assert(ghch);
      assert(ghch->up() == &ghost );

      // set element pointer and local face number 
      info.setGhostPair( ghostpair_STI( ghch, neighbour.second ) , count );

      ++count ;
    }
  }
}
 
template < class A, class X, class MX >
void Hbnd4PllInternal < A, X, MX > :: HbndPll ::
removeDescendents( helement_STI & elem )
{
  elem.resetRefinementRequest(); 

  // check all children first 
  for( helement_STI * child = elem.down(); child; child = child->next() )
  { 
    // if child is not leaf coarse childs first 
    if( ! child->leaf() )
      removeDescendents( *child );
    
    // if something went wrong, return ghosts are removed later 
    if( ! child->leaf () ) return ;
    
    // mark child for coarsening 
    child->tagForGlobalCoarsening();
  } 
    
  // if element is not already leaf call coarse 
  if( ! elem.leaf () )
  {
    elem.coarse();
  }
}
 
template < class A, class X, class MX >
void Hbnd4PllInternal < A, X, MX > :: HbndPll ::  coarseGhost ()
{
  if(_ghostPair.first)
  {
    helement_STI & ghost = (*_ghostPair.first);
    if( ghost.leaf() ) return ;

    // remove all descendents if possible 
    removeDescendents( ghost );
  }

}

////////////////////////////////////////////////////////////////////
//
//  --HbndPllMacro
//
////////////////////////////////////////////////////////////////////
template < class A, class X, class MX >
const MacroGhostInfo_STI* Hbnd4PllInternal < A, X, MX > ::
HbndPllMacro :: buildGhostCell(ObjectStream& os, int fce)
{
  assert( _gm == 0 );
  int code = MacroGridMoverIF :: ENDMARKER ;
  os.readObject (code);
  assert( code == MacroGridMoverIF :: HBND4INT );

  {
    int bfake, v [4] = {-1,-1,-1,-1};

    os.readObject (bfake) ;
#ifndef NDEBUG 
    Gitter :: hbndseg :: bnd_t b = (Gitter :: hbndseg :: bnd_t) bfake;
    assert( b == Gitter :: hbndseg :: closure );
#endif

    os.readObject (v[0]) ;
    os.readObject (v[1]) ;
    os.readObject (v[2]) ;
    os.readObject (v[3]) ;

    int readPoint = 0; 
    os.readObject( readPoint ); 
    
    // read ghost information 
    MacroGhostInfoHexa* ghInfo = 0;
    if( readPoint == MacroGridMoverIF :: POINTTRANSMITTED ) 
    {
      // read data from stream 
      ghInfo = new MacroGhostInfoHexa( os );
    }

    // create macro ghost cell     
    {
      assert( ghInfo );

      myhface4_t * f = this->myhface4(0);
      assert( f );

      _gm = new MacroGhostHexa( _mgb , ghInfo, f );
      this->setGhost ( _gm->getGhost() );
    }
  }

  assert( _gm );
  return _gm->getGhostInfo();
}


// template instantiation 
typedef GitterBasis :: Objects :: Hbnd4Default Hbnd4DefaultType;
template class Hbnd4PllInternal < Hbnd4DefaultType ,
                                  BndsegPllBaseXClosure < Hbnd4DefaultType > ,
                                  BndsegPllBaseXMacroClosure < Hbnd4DefaultType > > ;

// from serial part with different template argument 
template class Hedge1Top< GitterBasisPll :: ObjectsPll :: Hedge1EmptyPll >;
template class Hface4Top< GitterBasisPll :: ObjectsPll :: Hface4EmptyPll >;
template class HexaTop< GitterBasisPll :: ObjectsPll :: HexaEmptyPll >;
template class Periodic4Top < GitterBasisPll :: ObjectsPll :: Periodic4EmptyPll >;

